@@ -25,21 +25,25 @@
 #include "mm-modem-huawei-gsm.h"
 #include "mm-modem-gsm-network.h"
 #include "mm-modem-gsm-card.h"
+#include "mm-modem-gsm-ussd.h"
 #include "mm-errors.h"
 #include "mm-callback-info.h"
 #include "mm-at-serial-port.h"
 #include "mm-serial-parsers.h"
 #include "mm-log.h"
+#include "mm-utils.h"
 
 static void modem_init (MMModem *modem_class);
 static void modem_gsm_network_init (MMModemGsmNetwork *gsm_network_class);
 static void modem_gsm_card_init (MMModemGsmCard *gsm_card_class);
+static void modem_gsm_ussd_init (MMModemGsmUssd *ussd_class);
 
 G_DEFINE_TYPE_EXTENDED (MMModemHuaweiGsm, mm_modem_huawei_gsm, MM_TYPE_GENERIC_GSM, 0,
                         G_IMPLEMENT_INTERFACE (MM_TYPE_MODEM, modem_init)
                         G_IMPLEMENT_INTERFACE (MM_TYPE_MODEM_GSM_NETWORK, modem_gsm_network_init)
-                        G_IMPLEMENT_INTERFACE (MM_TYPE_MODEM_GSM_CARD, modem_gsm_card_init))
-
+                        G_IMPLEMENT_INTERFACE (MM_TYPE_MODEM_GSM_CARD, modem_gsm_card_init)
+                        G_IMPLEMENT_INTERFACE (MM_TYPE_MODEM_GSM_USSD, modem_gsm_ussd_init)
+)
 
 #define MM_MODEM_HUAWEI_GSM_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), MM_TYPE_MODEM_HUAWEI_GSM, MMModemHuaweiGsmPrivate))
 
@@ -171,6 +175,11 @@ set_allowed_mode_done (MMAtSerialPort *port,
 {
     MMCallbackInfo *info = (MMCallbackInfo *) user_data;
 
+    /* If the modem has already been removed, return without
+     * scheduling callback */
+    if (mm_callback_info_check_modem_removed (info))
+        return;
+
     if (error)
         info->error = g_error_copy (error);
 
@@ -232,11 +241,18 @@ get_allowed_mode_done (MMAtSerialPort *port,
                        gpointer user_data)
 {
     MMCallbackInfo *info = (MMCallbackInfo *) user_data;
-    MMModemHuaweiGsm *self = MM_MODEM_HUAWEI_GSM (info->modem);
+    MMModemHuaweiGsm *self;
     int mode_a, mode_b, u1, u2;
     guint32 band;
     MMModemGsmAllowedMode mode = MM_MODEM_GSM_ALLOWED_MODE_ANY;
 
+    /* If the modem has already been removed, return without
+     * scheduling callback */
+    if (mm_callback_info_check_modem_removed (info))
+        return;
+
+    self = MM_MODEM_HUAWEI_GSM (info->modem);
+
     if (error)
         info->error = g_error_copy (error);
     else if (parse_syscfg (self, response->str, &mode_a, &mode_b, &band, &u1, &u2, &mode))
@@ -271,8 +287,16 @@ set_band_done (MMAtSerialPort *port,
                gpointer user_data)
 {
     MMCallbackInfo *info = (MMCallbackInfo *) user_data;
-    MMModemHuaweiGsm *self = MM_MODEM_HUAWEI_GSM (info->modem);
-    MMModemHuaweiGsmPrivate *priv = MM_MODEM_HUAWEI_GSM_GET_PRIVATE (self);
+    MMModemHuaweiGsm *self;
+    MMModemHuaweiGsmPrivate *priv;
+
+    /* If the modem has already been removed, return without
+     * scheduling callback */
+    if (mm_callback_info_check_modem_removed (info))
+        return;
+
+    self = MM_MODEM_HUAWEI_GSM (info->modem);
+    priv = MM_MODEM_HUAWEI_GSM_GET_PRIVATE (self);
 
     if (error)
         info->error = g_error_copy (error);
@@ -321,10 +345,17 @@ get_band_done (MMAtSerialPort *port,
                gpointer user_data)
 {
     MMCallbackInfo *info = (MMCallbackInfo *) user_data;
-    MMModemHuaweiGsm *self = MM_MODEM_HUAWEI_GSM (info->modem);
+    MMModemHuaweiGsm *self;
     int mode_a, mode_b, u1, u2;
     guint32 band;
 
+    /* If the modem has already been removed, return without
+     * scheduling callback */
+    if (mm_callback_info_check_modem_removed (info))
+        return;
+
+    self = MM_MODEM_HUAWEI_GSM (info->modem);
+
     if (error)
         info->error = g_error_copy (error);
     else if (parse_syscfg (self, response->str, &mode_a, &mode_b, &band, &u1, &u2, NULL)) {
@@ -410,6 +441,11 @@ get_act_request_done (MMAtSerialPort *port,
     char *str;
     int srv_stat = 0;
 
+    /* If the modem has already been removed, return without
+     * scheduling callback */
+    if (mm_callback_info_check_modem_removed (info))
+        return;
+
     if (error) {
         info->error = g_error_copy (error);
         goto done;
@@ -504,6 +540,11 @@ send_huawei_cpin_done (MMAtSerialPort *port,
     char *str = NULL;
     guint32 num = 0;
 
+    /* If the modem has already been removed, return without
+     * scheduling callback */
+    if (mm_callback_info_check_modem_removed (info))
+        return;
+
     if (error) {
         info->error = g_error_copy (error);
         goto done;
@@ -710,12 +751,6 @@ do_enable_power_up_done (MMGenericGsm *gsm,
 
 /*****************************************************************************/
 
-typedef struct {
-    MMModem *modem;
-    MMModemFn callback;
-    gpointer user_data;
-} DisableInfo;
-
 static void
 disable_unsolicited_done (MMAtSerialPort *port,
                           GString *response,
@@ -723,12 +758,29 @@ disable_unsolicited_done (MMAtSerialPort *port,
                           gpointer user_data)
 
 {
-    MMModem *parent_modem_iface;
-    DisableInfo *info = user_data;
+    MMCallbackInfo *info = (MMCallbackInfo *) user_data;
 
-    parent_modem_iface = g_type_interface_peek_parent (MM_MODEM_GET_INTERFACE (info->modem));
-    parent_modem_iface->disable (info->modem, info->callback, info->user_data);
-    g_free (info);
+    /* If the modem has already been removed, return without
+     * scheduling callback */
+    if (mm_callback_info_check_modem_removed (info))
+        return;
+
+    /* Ignore all errors */
+    mm_callback_info_schedule (info);
+}
+
+static void
+invoke_call_parent_disable_fn (MMCallbackInfo *info)
+{
+    /* Note: we won't call the parent disable if info->modem is no longer
+     * valid. The invoke is called always once the info gets scheduled, which
+     * may happen during removed modem detection. */
+    if (info->modem) {
+        MMModem *parent_modem_iface;
+
+        parent_modem_iface = g_type_interface_peek_parent (MM_MODEM_GET_INTERFACE (info->modem));
+        parent_modem_iface->disable (info->modem, (MMModemFn)info->callback, info->user_data);
+    }
 }
 
 static void
@@ -737,12 +789,12 @@ disable (MMModem *modem,
          gpointer user_data)
 {
     MMAtSerialPort *primary;
-    DisableInfo *info;
+    MMCallbackInfo *info;
 
-    info = g_malloc0 (sizeof (DisableInfo));
-    info->callback = callback;
-    info->user_data = user_data;
-    info->modem = modem;
+    info = mm_callback_info_new_full (modem,
+                                      invoke_call_parent_disable_fn,
+                                      (GCallback)callback,
+                                      user_data);
 
     primary = mm_generic_gsm_get_at_port (MM_GENERIC_GSM (modem), MM_PORT_TYPE_PRIMARY);
     g_assert (primary);
@@ -826,6 +878,42 @@ out:
     return !!port;
 }
 
+/* Encode to packed GSM - this is what Huawei supports on all known models */
+static char*
+ussd_encode (MMModemGsmUssd *self, const char* command, guint *scheme)
+{
+    char *hex;
+    guint8 *gsm, *packed;
+    guint32 len = 0, packed_len = 0;
+
+    *scheme = MM_MODEM_GSM_USSD_SCHEME_7BIT;
+    gsm = mm_charset_utf8_to_unpacked_gsm (command, &len);
+    packed = gsm_pack (gsm, len, 0, &packed_len);
+    hex = utils_bin2hexstr (packed, packed_len);
+    g_free (packed);
+    g_free (gsm);
+
+    return hex;
+}
+
+/* Unparse packed gsm to utf8 */
+static char*
+ussd_decode (MMModemGsmUssd *self, const char* reply, guint scheme)
+{
+    char *bin, *utf8;
+    guint8 *unpacked;
+    gsize bin_len;
+    guint32 unpacked_len;
+
+    bin = utils_hexstr2bin (reply, &bin_len);
+    unpacked = gsm_unpack ((guint8*)bin, bin_len, 0, &unpacked_len);
+    utf8 = (char*)mm_charset_gsm_unpacked_to_utf8 (unpacked, unpacked_len);
+
+    g_free (bin);
+    g_free (unpacked);
+    return utf8;
+}
+
 /*****************************************************************************/
 
 static void
@@ -853,6 +941,13 @@ mm_modem_huawei_gsm_init (MMModemHuaweiGsm *self)
 {
 }
 
+static void
+modem_gsm_ussd_init (MMModemGsmUssd *ussd_class)
+{
+    ussd_class->encode = ussd_encode;
+    ussd_class->decode = ussd_decode;
+}
+
 static void
 mm_modem_huawei_gsm_class_init (MMModemHuaweiGsmClass *klass)
 {
