@@ -30,6 +30,10 @@ static gboolean impl_manager_enumerate_devices (MMManager *manager,
                                                 GPtrArray **devices,
                                                 GError **err);
 
+static gboolean impl_manager_set_logging (MMManager *manager,
+                                          const char *level,
+                                          GError **error);
+
 #include "mm-manager-glue.h"
 
 G_DEFINE_TYPE (MMManager, mm_manager, G_TYPE_OBJECT)
@@ -912,12 +916,26 @@ handle_uevent (GUdevClient *client,
 	/* We only care about tty/net devices when adding modem ports,
 	 * but for remove, also handle usb parent device remove events
 	 */
-	if ((!strcmp (action, "add") || !strcmp (action, "move")) && strcmp (subsys, "usb") !=0 )
+	if (   (!strcmp (action, "add") || !strcmp (action, "move") || !strcmp (action, "change"))
+        && (strcmp (subsys, "usb") != 0))
 		device_added (self, device);
 	else if (!strcmp (action, "remove"))
 		device_removed (self, device);
 }
 
+static gboolean
+impl_manager_set_logging (MMManager *manager,
+                          const char *level,
+                          GError **error)
+{
+	if (mm_log_set_level (level, error)) {
+		mm_info ("logging: level '%s'", level);
+		return TRUE;
+	}
+	return FALSE;
+}
+
+
 void
 mm_manager_start (MMManager *manager)
 {
